(*
XXX

Number of packets/weight/direction

- Packages
- Drop/Pick
- Evaluation function

- The time limit should depend on some hotness factor
  (higher if we are closer to other robots or stuck)

Smart: push us (we are not at the expected location)
Need direction vector

Drop/pick
*)

let m_depth = ref 0
let space_of_depth d =
  String.make (2 * (2 * !m_depth + 2 - d)) ' '

type robot =
  { id : int;  (* 0 is me *)
    mutable x : int;
    mutable y : int;
    x0 : int;
    y0 : int;
    mutable disallowed : int;
    mutable killed : bool;
    smart : bool }

type score = (float * int * int)

type state =
  { all_robots : robot list;
    dists : int array;
    values : (int, score) Hashtbl.t;
    mutable package : int;
    mutable weight : int;
    mutable score : int;
    mutable use_dist : bool }

(****)

let get_distance state x y =
  if state.use_dist then
    Distance.get_distance state.dists x y
  else
    0

let eval state =
(*
10000 per killed robot
50 per explored homes
progress (distance + 5 * score + 2 * package load variation)
*)
  let me = List.find (fun r -> r.id = 0) state.all_robots in
  let deads = ref 0 in
  List.iter (fun r -> if r.killed then incr deads) state.all_robots;
  let score =
    0.
      +. 1000. *. float !deads
      +. float (get_distance state me.x0 me.y0)
      -. float (get_distance state me.x me.y)
(*      +. 10. *. float state.weight*)
      +. float state.weight
      +. float state.score
  in
  if !Debug.ab then Format.eprintf "%sScore: %f@." (space_of_depth 0) score;
  score

(****)

let timer = Time.make_timer ()

(****)

let hash i r = i + 137 * r

let hash_state depth state =
  List.fold_right
    (fun r rem ->
       if r.killed then
         hash 0 rem
       else begin
         hash 1 (
         hash r.x (
         hash r.y (
         rem)))
       end)
    state.all_robots (hash state.weight (hash state.score depth))

let state_value state hash =
  Hashtbl.find state.values hash

let state_add_value state hash v =
  Hashtbl.add state.values hash v

(****)

let actions =
  [| (-1, 0,  1);
     (1, 0,  2);
     (0, -1,  4);
     (0, 1, 8);
     (0, 0, 16) |]

let conflict robot x y = abs (x - robot.x) + abs (y - robot.y) <= 1
let conflict_after_move robot x y = abs (x - robot.x) + abs (y - robot.y) = 1

let action_of_dir =
  [| -1;  2; -1;
      0;  4;  1;
     -1;  3; -1 |]

let disallowed_move robot action = (robot.disallowed land action) <> 0

(* Hash of current state. If same hash, we assume same state. *)

let mAX_DEPTH = 30
let stack = Array.make (6 * mAX_DEPTH * 4 * 3) 0
let stack2 = Array.make (6 * mAX_DEPTH * 4 * 3) false
let top = ref 0
let top2 = ref 0

let push_robot robot =
  let t = !top in
  assert (t >= 0 && t + 2 <= Array.length stack);
  stack.(t + 0) <- robot.x;
  stack.(t + 1) <- robot.y;
  top := t + 2;
  stack2.(!top2) <- robot.killed;
  incr top2

let pop_robot robot =
  let t = !top - 2 in
  assert (t >= 0 && t + 2 <= Array.length stack);
  robot.x <- stack.(t + 0);
  robot.y <- stack.(t + 1);
  top := t;
  decr top2;
  robot.killed <- stack2.(!top2)

let rec push_robots_rec robots =
  match robots with
    []       -> ()
  | r :: rem -> push_robot r; push_robots_rec rem

let push_robots state =
  let t = !top in
  stack.(t + 0) <- state.package;
  stack.(t + 1) <- state.weight;
  top := t + 2;
  stack2.(!top2) <- state.use_dist;
  incr top2;
  push_robots_rec state.all_robots

let rec pop_robots_rec robots =
  match robots with
    []       -> ()
  | r :: rem -> pop_robots_rec rem; pop_robot r

let pop_robots state =
  pop_robots_rec state.all_robots;
  decr top2;
  state.use_dist <- stack2.(!top2);
  let t = !top - 2 in
  state.package <- stack.(t + 0);
  state.weight <- stack.(t + 1);
  top := t

let fail = -1e6
let success = 1e10
let really_fail = -1e10
let really_bad v = v < fail /. 2.
let clearly_worse v1 v2 = v1 <= v2 -. 500.

exception Pruned

let add_score depth alpha beta worse v =
  if !worse > v then begin
    worse := v;
    if !beta > v then beta := v;
    if v <= alpha then raise Pruned
  end;
  if !Debug.ab then
    Format.eprintf "%s%g %g %g %g@."
      (space_of_depth (2 * depth + 1)) alpha !beta !worse v

exception Wall

let rec move_robot_rec depth robot state dx dy worse =
  let x' = robot.x + dx in
  let y' = robot.y + dy in
  match State.get_contents x' y' with
    '~' ->
      robot.killed <- true;
      if robot.id = 0 then begin
        worse := fail /. 2. ** (float (!m_depth - depth - 1));
        raise Pruned
      end
  | '#' ->
      raise Wall
  | _   ->
      List.iter
        (fun r ->
           if not (r.killed) && x' = r.x && y' = r.y then begin
             if r.id = 0 then begin
               state.weight <-
                 truncate (0.5 +. float state.weight
                             *. float (state.package - 1)
                             /. float state.package);
               state.package <- state.package - 1
             end;
             move_robot_rec depth r state dx dy worse
           end)
        state.all_robots;
      robot.x <- x'; robot.y <- y'

let rec move_robot depth robot state dx dy worse =
  if dx = 0 && dy = 0 then begin
    (* XXX Pick/drop *)
(*
Destination: drop
    ===> nouveaux objectifs (+ valeur de l'objectif)
Home: pick
    ===> nouvel objectif
*)
  end else
    try
      move_robot_rec depth robot state dx dy worse
    with Wall ->
      ()

let swap moves i j =
  let m = moves.(i) in
  moves.(i) <- moves.(j);
  moves.(j) <- m

let order_moves () =
  let moves = [|0;1;2;3;4|] in
  for i = 0 to 4 do
    let j = i + Random.int (5 - i) in
    swap moves i j
  done;
  moves

let order_my_moves state =
  let moves = order_moves () in
  let me = List.find (fun r -> r.id = 0) state.all_robots in
  let x = me.x in
  let y = me.y in
  let d = get_distance state x y in
  let best = ref 0 in
  let worse = ref 5 in
  let pos = ref 0 in
  while !pos < !worse do
    let (dx, dy, _) = actions.(moves.(!pos)) in
    let d' = get_distance state (x + dx) (y + dy) in
    if d' < d then begin
      swap moves !best !pos;
      incr best
    end else if d' > d then begin
      decr worse;
      swap moves !worse !pos
    end;
    incr pos
  done;
  moves

let combine_scores depth alpha beta worse bid smart v1 v2 =
  let v = ref 0. in
  let b = ref 1 in
  if v1 <= v2 then begin
v := v1;
(*    v := 0.8 *. v1 +. 0.2 *. v2;*)
    if clearly_worse v1 v2 then begin
      if really_bad v1 then begin
        b := -1000
      end else begin
        b := -500
      end
    end else begin
      if not (really_bad v1 || smart) then v := v2(*0.8 *. v2 +. 0.2 *. v1*);
      b := -1
    end
  end else begin
    v := v2;
(*
    v := 0.8 *. v2 +. 0.2 *. v1;
*)
    if clearly_worse v2 v1 then begin
      if really_bad v2 then
        b := 1000
      else
        b := 500
    end else begin
      if not (really_bad v2 || smart) then v := v1(*0.8 *. v1 +. 0.2 *. v2*);
      b := 2
    end
  end;
  if !worse > !v then bid := !b;
  if !Debug.ab then
    Format.eprintf "Combine: %g %g %g %d@." v1 v2 !v !bid;
  add_score depth alpha beta worse !v


(* We try to minimize our score
   We stop as soon as we find a score worse than alpha *)
let rec try_moves depth rem_robots state alpha beta worse =
  match rem_robots with
    [] ->
      let (score, _, _) = handle_robots depth state alpha !beta in
      add_score depth alpha beta worse score
  | robot :: rem_robots ->
      if robot.killed then
        try_moves depth rem_robots state alpha beta worse
      else
        let bid = ref 0 in
        let moves = order_moves () in
        for i = 0 to 4 do
          let n = moves.(i) in
          try_move depth robot rem_robots state n alpha beta worse bid
        done

and try_move depth robot rem_robots state n alpha beta worse bid =
  if !Debug.ab then
    Format.eprintf "Try_move %d %d@." depth (List.length rem_robots);
  let (dx, dy, action) = actions.(n) in
  let x' = robot.x + dx in
  let y' = robot.y + dy in
  if
    not (State.allowed_location x' y')
      ||
    disallowed_move robot action
  then begin
    if robot.id = 0 then add_score depth alpha beta worse really_fail
  end else
    find_conflicts
      depth robot rem_robots rem_robots state x' y' n alpha beta worse bid

and find_conflicts depth
    robot confl_robots rem_robots state x' y' n alpha beta worse bid =
  match confl_robots with
    [] ->
      push_robots state;
      begin try
        let (dx, dy, action) = actions.(n) in
        move_robot depth robot state dx dy worse;
        try_moves depth rem_robots state alpha beta worse;
      with Pruned ->
        pop_robots state;
        raise Pruned
      end;
      pop_robots state;
  | robot2 :: confl_robots ->
  if !Debug.ab then
    Format.eprintf "x: %d y: %d x2: %d y2: %d x' %d y' %d c1: %b c2: %b c3: %b@." robot.x robot.y robot2.x robot2.y x' y' (conflict_after_move robot2 robot.x robot.y) (conflict_after_move robot2 x' y') (conflict_after_move robot robot2.x robot2.y);
      if x' = robot2.x && y' = robot2.y then begin
        let n2 =
          action_of_dir.((robot.x - robot2.x) + 3 * (robot.y - robot2.y) + 4)
        in
        let (dx, dy, action) = actions.(n) in
        let rem_robots' =
          List.filter (fun r -> r.id <> robot2.id) rem_robots in
        (* Robot 1 move before robot2 *)
        push_robots state;
        let worse1 = ref !worse in
        begin try
          move_robot depth robot state dx dy worse1;
          let beta = ref !beta in
          try_moves depth rem_robots' state alpha beta worse1
        with Pruned -> () end;
        pop_robots state;
        (* Robot 2 move before robot1 *)
        let worse2 = ref !worse in
        begin try
          let beta = ref !beta in
          for i = 0 to 4 do
            push_robots state;
            begin try
              let (dx2, dy2, action2) = actions.(i) in
              move_robot depth robot2 state dx2 dy2 worse2;
              if i <> n2 then
                move_robot depth robot state dx dy worse2;
              try_moves depth rem_robots' state alpha beta worse2;
            with Pruned ->
              pop_robots state;
              raise Pruned
            end;
            pop_robots state
          done
        with Pruned -> () end;
        combine_scores depth alpha beta worse bid robot2.smart !worse1 !worse2
      end else if
        conflict_after_move robot2 robot.x robot.y
          ||
        conflict_after_move robot2 x' y'
      then begin
        let (x'', y'') =
          if conflict_after_move robot2 robot.x robot.y then
            (robot.x, robot.y)
          else
            (x', y')
        in
        let n2 =
          action_of_dir.((x'' - robot2.x) + 3 * (y'' - robot2.y) + 4)
        in
        let (dx, dy, action) = actions.(n) in
        let (dx2, dy2, action2) = actions.(n2) in
        let rem_robots' =
          List.filter (fun r -> r.id <> robot2.id) rem_robots in
        (* Robot 1 move before robot2 *)
        push_robots state;
        let worse1 = ref !worse in
        begin try
          move_robot depth robot state dx dy worse1;
          move_robot depth robot2 state dx2 dy2 worse1;
          let beta = ref !beta in
          try_moves depth rem_robots' state alpha beta worse1
        with Pruned -> () end;
        pop_robots state;
        (* Robot 2 move before robot1 *)
        push_robots state;
        let worse2 = ref !worse in
        begin try
          let move1 = conflict_after_move robot2 x' y' in
          move_robot depth robot2 state dx2 dy2 worse2;
          if move1 then
            move_robot depth robot state dx dy worse2;
          let beta = ref !beta in
          try_moves depth rem_robots' state alpha beta worse2;
        with Pruned -> () end;
        pop_robots state;
        combine_scores depth alpha beta worse bid robot2.smart !worse1 !worse2;
        (* Other moves *)
        let disallowed = robot2.disallowed in
        robot2.disallowed <- robot2.disallowed lor action_of_dir.(n);
        push_robots state;
        begin try
          let (dx, dy, action) = actions.(n) in
          move_robot depth robot state dx dy worse;
          try_moves depth rem_robots state alpha beta worse;
        with Pruned ->
          pop_robots state;
          robot2.disallowed <- disallowed;
          raise Pruned
        end;
        pop_robots state;
        robot2.disallowed <- disallowed
      end else
        find_conflicts depth
          robot confl_robots rem_robots state x' y' n alpha beta worse bid

(* We try to maximize our score
   We stop as soon as we find a score better than beta *)
and handle_robots depth state alpha beta =
  if !Debug.ab then Format.eprintf "%d@." depth;
  if !Debug.ab then begin
    Format.eprintf "Locs: ";
    List.iter (fun r -> Format.eprintf "%d %d -" r.x r.y) state.all_robots;
    Format.eprintf "@."
  end;
  if Time.timeout timer then begin
    if !Debug.ab then
      Format.eprintf "Timeout@.";
    (really_fail, 0, 1)
  end else
    let hash = hash_state depth state in
    try
      let res=
        state_value state hash
      in
      if !Debug.ab then
        Format.eprintf "Hit!@.";
      res
    with Not_found ->
      let me = List.find (fun r -> r.id = 0) state.all_robots in
      if
        state.use_dist &&
        me.id = 0 && get_distance state me.x me.y = 0
      then begin
        state.score <- 20 * depth + 1
            + get_distance state me.x0 me.y0
            - get_distance state me.x me.y;
        state.use_dist <- false
      end;
      let res =
        if depth = 0 then begin
          (* XXX Evaluation function *)
          (eval state, 0, 1)
        end else begin
          let alpha = ref alpha in
          let best = ref really_fail in
          let best_move = ref 0 in
          let best_bid = ref 1 in
          try
            let moves = order_my_moves state in
            for i = 0 to 4 do
              let n = moves.(i) in
              let worse = ref success in
              let bid = ref 1 in
              begin try
                let beta = ref beta in
                try_move
                  (depth - 1)
                  (List.hd state.all_robots) (List.tl state.all_robots)
                  state n !alpha beta worse bid;
              with Pruned -> () end;
              if !best < !worse then begin
                best := !worse;
                best_move := n;
                best_bid := !bid;
                if !alpha < !worse then alpha := !worse;
                if !worse >= beta then raise Pruned
              end;
              if !Debug.ab then
                Format.eprintf "%s%g %g %g %g@."
                  (space_of_depth (2 * depth)) !alpha beta !best !worse
            done;
            (!best, !best_move, !best_bid)
          with Pruned ->
            (!best, !best_move, !best_bid)
        end
      in
      state_add_value state hash res;
      res

let best_move max_depth state =
  Time.start_timer timer 0.5;
  let best_move = ref 0 in
  let best_bid = ref 1 in
  let score = ref 0. in
  let depth = ref 1 in
  while !depth < max_depth && (!depth = 1 || not (Time.timeout timer)) do
    Hashtbl.clear state.values;
    m_depth := !depth;
    let (best, move, bid) = handle_robots !depth state really_fail success in
    if !depth = 1 || not (Time.timeout timer) then begin
      score := best;
      best_move := move;
      best_bid := bid
    end;
    incr depth
  done;
  if !Communication.print_map then
    Format.eprintf "Move: %d  Bid: %d  Score: %f  Depth: %d@."
      !best_move !best_bid !score (!depth - 1);
  (!best_move, !best_bid)

let find_move d scope robots =
  let max_depth = 2 + min 18 scope in
  let (x, y) = State.my_location () in
  let id = ref 1 in
  let packages = !(State.get_robot_packages !State.my_id) in
  let weight = ref 0 in
  List.iter
    (fun id ->
       weight := !weight + (Hashtbl.find State.package_dests id).State.weight)
    packages;
  let state =
    { dists = d;
      all_robots =
        { id = 0; x = x; y = y; x0 = x; y0 = y;
          disallowed = 0; killed = false; smart = true } ::
        List.map
          (fun r ->
             incr id;
             let (x, y) = Hashtbl.find State.robot_locations r in
             { id = !id; x = x; y = y; x0 = x; y0 = y;
               disallowed = 0; killed = false; smart = false })
          robots;
      values = Hashtbl.create 101;
      package = List.length packages; weight = !weight; score = 0;
      use_dist = true }
  in
  let (move, bid) = best_move max_depth state in
  (begin match move with
     0 -> 'w'
   | 1 -> 'e'
   | 2 -> 's'
   | 3 -> 'n'
   | 4 -> 'd'
   | _ -> assert false
   end,
   (if bid < -500 then - min 500 (max 1 (!State.money / 2)) else
    if bid > 500 then min 500 (max 1 (!State.money / 2)) else
    if bid > !State.money / 10 then max 1 !State.money / 10 else
    if bid < - !State.money /10 then - max 1 !State.money /10 else
    bid),
   [])
